from warnings import warn

from six import add_metaclass


class ServiceBaseMeta(type):

    """Metaclass that adds the `_name` attribute to any `statsd_dep.StatsD`
    object found in the attributes.
    """

    def __new__(cls, name, bases, attrs):
        warn(
            "Use of the `ServiceBaseMeta` metaclass is no longer needed."
            " `ServiceBaseMeta` is deprecated and will be removed in a future"
            " release of `nameko-statsd`.",
            DeprecationWarning
        )

        return type.__new__(cls, name, bases, attrs)


@add_metaclass(ServiceBaseMeta)
class ServiceBase(object):
    """Service base class. """
    pass



import warnings

from six import add_metaclass

from nameko_statsd.bases import ServiceBaseMeta


def test_metaclass():

    @add_metaclass(ServiceBaseMeta)
    class DummyService(object):

        def some_method(self):
            pass

    assert DummyService.__name__ == 'DummyService'


def test_metaclass_deprecated(recwarn):
    warnings.simplefilter("always")

    ServiceBaseMeta('DummyService', (object,), {})

    assert len(recwarn) == 1
    warn = recwarn.pop()
    assert str(warn.message) == (
        "Use of the `ServiceBaseMeta` metaclass is no longer needed."
        " `ServiceBaseMeta` is deprecated and will be removed in a future"
        " release of `nameko-statsd`."
    )



from mock import Mock, call, patch
import pytest

from nameko_statsd.statsd_dep import LazyClient, Protocols


class TestLazyClient(object):

    """Test the `LazyClient` class features. """

    @pytest.fixture(autouse=True)
    def stats_client_cls(self):
        with patch('nameko_statsd.statsd_dep.StatsClient') as sc:
            yield sc

    @pytest.fixture(autouse=True)
    def stats_client_cls_tcp(self):
        with patch('nameko_statsd.statsd_dep.TCPStatsClient') as sc:
            yield sc


class TestLazyClientUDP(TestLazyClient):

    @pytest.fixture
    def stats_config(self, stats_config):
        return stats_config['STATSD']['test'].copy()

    def test_init(self, stats_client_cls, stats_client_cls_tcp, stats_config):
        lc = LazyClient(**stats_config)

        assert lc.config == dict(
            host='statsd.host',
            port=1234,
            prefix='statsd.prefix',
            maxudpsize=1024,
        )
        assert lc.enabled
        assert lc._client is None
        assert stats_client_cls.call_count == 0
        assert stats_client_cls_tcp.call_count == 0

    @pytest.mark.parametrize('proto', ('UPD', 'upd', 'IP', 'TPC', 'tPc'))
    def test_init_invalid_protocol(
        self, stats_client_cls, stats_client_cls_tcp, stats_config, proto
    ):
        with pytest.raises(ValueError) as err:
            LazyClient(protocol=proto, **stats_config)

        assert err.match('Invalid protocol')

    def test_client_default_protocol(
        self, stats_client_cls, stats_client_cls_tcp, stats_config
    ):
        lc = LazyClient(**stats_config)

        client = lc.client  # this is a property

        assert lc.protocol is Protocols.udp

        assert stats_client_cls.call_args_list == [
            call(
                host='statsd.host',
                port=1234,
                prefix='statsd.prefix',
                maxudpsize=1024,
            )
        ]
        assert client == stats_client_cls.return_value
        assert stats_client_cls_tcp.call_count == 0

    @pytest.mark.parametrize('proto', ('udp', 'UDP', 'UdP', 'uDp'))
    def test_client_udp(
        self, stats_client_cls, stats_client_cls_tcp, stats_config, proto
    ):
        lc = LazyClient(protocol=proto, **stats_config)

        client = lc.client  # this is a property

        assert lc.protocol is Protocols.udp

        assert stats_client_cls.call_args_list == [
            call(
                host='statsd.host',
                port=1234,
                prefix='statsd.prefix',
                maxudpsize=1024,
            )
        ]
        assert client == stats_client_cls.return_value
        assert stats_client_cls_tcp.call_count == 0


class TestLazyClientTCP(TestLazyClient):

    @pytest.fixture
    def stats_config(self, stats_config):
        return stats_config['STATSD']['test-tcp'].copy()

    def test_init(
        self, stats_client_cls, stats_client_cls_tcp, stats_config
    ):
        lc = LazyClient(**stats_config)

        assert lc.config == dict(
            host='tcp.statsd.host',
            port=4321,
            prefix='tcp.statsd.prefix',
            timeout=5,
        )
        assert lc.enabled
        assert lc._client is None
        assert stats_client_cls.call_count == 0
        assert stats_client_cls_tcp.call_count == 0

    @pytest.mark.parametrize('proto', ('tcp', 'TCP', 'TcP', 'tCp'))
    def test_client(
        self, stats_client_cls, stats_client_cls_tcp, stats_config, proto
    ):
        del stats_config['protocol']
        lc = LazyClient(protocol=proto, **stats_config)

        client = lc.client  # this is a property

        assert lc.protocol is Protocols.tcp

        assert stats_client_cls_tcp.call_args_list == [
            call(
                host='tcp.statsd.host',
                port=4321,
                prefix='tcp.statsd.prefix',
                timeout=5,
            )
        ]
        assert client == stats_client_cls_tcp.return_value
        assert stats_client_cls.call_count == 0


class TestStatMethods(TestLazyClient):

    @pytest.fixture(params=['test', 'test-tcp'])
    def stats_config(self, request, stats_config):
        return stats_config['STATSD'][request.param].copy()

    @pytest.mark.parametrize('method', [
        'incr', 'decr', 'gauge', 'set', 'timing', 'timer', 'pipeline',
    ])
    def test_passthrough_methods(self, method, stats_config):
        lc = LazyClient(**stats_config)

        delegate = getattr(lc, method)

        # make the call
        delegate(1, 2, 3, 'Darth Vader', obi_wan='Kenobi')

        # verify it's a passthrough
        assert getattr(lc.client, method).call_args_list == [
            call(1, 2, 3, 'Darth Vader', obi_wan='Kenobi')
        ]

    @pytest.mark.parametrize('method', [
        'incr', 'decr', 'gauge', 'set', 'timing', 'timer', 'pipeline',
    ])
    def test_passthrough_disabled(
        self, method, stats_config
    ):
        stats_config['enabled'] = False
        lc = LazyClient(**stats_config)

        delegate = getattr(lc, method)

        # make the call
        delegate(1, 2, 3, 'Darth Vader', obi_wan='Kenobi')

        assert getattr(lc.client, method).call_args_list == []

    def test_getattr(self, stats_config):
        lc = LazyClient(**stats_config)

        with pytest.raises(AttributeError) as exc_info:
            lc.missing_method()

        assert exc_info.match(
            "'LazyClient' object has no attribute 'missing_method'"
        )

    def test_timer_contextmanager_enabled(
        self, stats_client_cls, stats_config
    ):
        lc = LazyClient(**stats_config)

        with lc.timer('stat', rate=2) as timer:
            pass

        assert timer is lc.client.timer.return_value.__enter__.return_value
        assert lc.client.timer.call_args_list == [call('stat', rate=2)]

    def test_timer_contextmanager_disabled(
        self, stats_client_cls, stats_config
    ):
        stats_config['enabled'] = False
        lc = LazyClient(**stats_config)

        with lc.timer('stat', rate=2) as timer:
            pass

        assert isinstance(timer, Mock)
        assert lc.client.timer.call_args_list == []

    def test_pipeline_contextmanager_enabled(
            self, stats_client_cls, stats_config
    ):
        lc = LazyClient(**stats_config)

        with lc.pipeline() as pipe:
            pipe.incr('foo')
            pipe.send()

        assert pipe is lc.client.pipeline.return_value.__enter__.return_value
        assert lc.client.pipeline.mock_calls == [
            call(),
            call().__enter__(),
            call().__enter__().incr('foo'),
            call().__enter__().send(),
            call().__exit__(None, None, None)
        ]

    def test_pipeline_contextmanager_disabled(
        self, stats_client_cls, stats_config
    ):
        stats_config['enabled'] = False
        lc = LazyClient(**stats_config)

        with lc.pipeline() as pipe:
            pipe.incr('foo')
            pipe.send()

        assert isinstance(pipe, Mock)
        assert lc.client.pipeline.mock_calls == []



from enum import Enum
from functools import wraps, partial
from mock import MagicMock
from warnings import warn

from nameko.extensions import DependencyProvider
from statsd import StatsClient, TCPStatsClient


class Protocols(Enum):
    tcp = 'tcp'
    udp = 'udp'


class LazyClient(object):

    """Provide an interface to `StatsClient` with a lazy client creation.
    """

    def __init__(self, **config):
        self.config = config
        self.enabled = config.pop('enabled')
        self._client = None

        protocol = self.config.pop('protocol', Protocols.udp.name)

        try:
            self.protocol = getattr(Protocols, protocol.lower())
        except AttributeError:
            raise ValueError(
                'Invalid protocol: {}'.format(protocol)
            )

    @property
    def client(self):
        if self._client is None:
            if self.protocol is Protocols.udp:
                self._client = StatsClient(**self.config)
            else:   # self.protocol is Protocols.tcp
                self._client = TCPStatsClient(**self.config)

        return self._client

    def __getattr__(self, name):
        if name in ('incr', 'decr', 'gauge', 'set', 'timing'):
            return partial(self._passthrough, name)
        else:
            message = "'{cls}' object has no attribute '{attr}'".format(
                cls=self.__class__.__name__, attr=name
            )
            raise AttributeError(message)

    def _passthrough(self, name, *args, **kwargs):
        if self.enabled:
            return getattr(self.client, name)(*args, **kwargs)

    def timer(self, *args, **kwargs):
        if self.enabled:
            return self.client.timer(*args, **kwargs)
        else:
            return MagicMock()

    def pipeline(self, *args, **kwargs):
        if self.enabled:
            return self.client.pipeline(*args, **kwargs)
        else:
            return MagicMock()


class StatsD(DependencyProvider):

    def __init__(self, key, name=None, *args, **kwargs):
        """
        Args:
            key (str): The key under the `STATSD` config dictionary.
            name (str): The name associated to the instance.
        """
        self._key = key

        if name is not None:
            warn(
                "The `name` argument to `StatsD` is no longer needed and has"
                " been deprecated.",
                DeprecationWarning
            )

        super(StatsD, self).__init__(*args, **kwargs)

    def get_dependency(self, worker_ctx):
        return LazyClient(**self.config)

    def setup(self):
        self.config = self.get_config()
        return super(StatsD, self).setup()

    def get_config(self):
        return self.container.config['STATSD'][self._key]

    def timer(self, *targs, **tkwargs):
        """Decorate a nameko service method.

        It can be applied to any instance method in a nameko service class,
        even to RPC or HTTP entrypoints.  If `dependency.enabled` is `False`
        this decorator is equivalent to a no-op.

        Args:
            *targs: Positional arguments to be given to `StatsClient.timer()`.
            *tkwargs: Keyword arguments to be given to `StatsClient.timer()`.

        Decorating a service like this:

        .. code-block:: python

            class MyService:

                statsd = StatsD('config_key')

                @rpc (or @http or even nothing)
                @statsd.timer('my_stat', rate=5)
                def method(...):
                    # method body

        is equivalent to the following:

        .. code-block:: python

            class MyService:

                statsd = StatsD('config_key')

                @rpc (or @http or even nothing)
                def method(...):
                    with self.statsd.client.timer('my_stat', rate=5):
                        # method body

        """
        def decorator(method):

            @wraps(method)
            def wrapper(svc, *args, **kwargs):
                # self.attr_name starts as None, then is set to the name of
                # the attribute when bind is called. Until then, the decorator
                # does nothing
                if self.attr_name is None:
                    return method(svc, *args, **kwargs)

                dependency = getattr(svc, self.attr_name)

                if dependency.enabled:
                    with dependency.client.timer(*targs, **tkwargs):
                        res = method(svc, *args, **kwargs)
                else:
                    res = method(svc, *args, **kwargs)

                return res

            return wrapper

        return decorator
